<?php
namespace Drupal\entity_classes_override\Form;

use Drupal\Core\Ajax\AjaxResponse;
use Drupal\Core\Ajax\ReplaceCommand;
use Drupal\Core\Cache\Cache;
use Drupal\Core\Plugin\Discovery\AnnotatedClassDiscovery;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\ContentEntityTypeInterface;
use Drupal\Core\Entity\EntityDisplayRepositoryInterface;
use Drupal\Core\Entity\EntityTypeBundleInfoInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\ConfigFormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\entity_classes_override\OverrideService;
use Symfony\Component\DependencyInjection\ContainerInterface;
class SettingsForm extends ConfigFormBase {
  private ?OverrideService $service = NULL;
  private array $configData;

  public function __construct(ConfigFactoryInterface $config_factory,OverrideService $overrideService) {
    parent::__construct($config_factory);
    $this->service = $overrideService;
    $this->configData = $this->config('entity_classes_override.settings')->getRawData();
  }
  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('config.factory'),
      $container->get('entity_classes_override.service')
    );
  }
  public function getFormId() {
    return 'entity_classes_override_settings_form';
  }
  protected function getEditableConfigNames()  {
    return ['entity_classes_override.settings'];
  }
  private function service(): OverrideService {
    if(isset($this->service) || is_null($this->service)) {
      $service = \Drupal::service('entity_classes_override.service');
      $this->service = $service;
      return $service;
    }
    return $this->service;
  }
  private function configData(){
    if(!isset($this->configData)){
      $this->configData = \Drupal::config('entity_classes_override.settings')->getRawData();
    }
    return $this->configData;
  }
  private function deleteButton($name): array  {
    return [
      '#type' => 'submit',
      '#value' => $this->t('Remove'),
      '#name' => "remove|$name",
    ];
  }
  public function buildForm(array $form, FormStateInterface $form_state) {
    $this->messenger()->addWarning($this->t("Bundle classes must extend the active entity class."));
    $form = parent::buildForm($form, $form_state);
    $form['#attributes']['novalidate'] = 'novalidate';
    $form['overrides'] = [
      '#type' => 'table',
      '#header' => [
        $this->t('Entity type'),
        $this->t("Override type"),
        $this->t("Original class"),
        $this->t("Current class"),
        $this->t("Remove")
      ]
    ];
    $x = 1;
    foreach($this->service()->contentEntityTypes as $contentEntityTypeId => $contentEntityType) {
      /** @var ContentEntityTypeInterface $contentEntityType */
      //check entity class
      if($contentEntityType->getClass() != $contentEntityType->getOriginalClass()) {
        $form['overrides'][$x]["entity_type"] = [
          '#markup' => $contentEntityTypeId,
        ];
        $form['overrides'][$x]["override"] = [
          '#markup' => $this->t("Entity class"),
        ];
        $form['overrides'][$x]["original"] = [
          '#markup' => $contentEntityType->getOriginalClass(),
        ];
        $form['overrides'][$x]["current"] = [
          '#markup' => $contentEntityType->getClass(),
        ];
        if(isset($this->configData()['entity_class'][$contentEntityTypeId][$contentEntityTypeId]) && $this->configData()['entity_class'][$contentEntityTypeId][$contentEntityTypeId] == $contentEntityType->getClass()) {
          $buttonName = [
            'entity_class',
            $contentEntityTypeId,
            $contentEntityTypeId,
            $contentEntityType->getClass()
          ];
          $form['overrides'][$x]["remove"] = $this->deleteButton(implode('|', $buttonName));
        }
        $x++;
      }
      //check bundle classes
      foreach($this->service()->bundleInfo->getBundleInfo($contentEntityTypeId) as $bundle => $bundleInfo) {
        if(isset($bundleInfo['class'])) {
          $form['overrides'][$x]["entity_type"] = [
            '#markup' => $contentEntityTypeId . ":" . $bundle,
          ];
          $form['overrides'][$x]["override"] = [
            '#markup' => $this->t("Entity bundle class"),
          ];
          $form['overrides'][$x]["original"] = [
            '#markup' => $contentEntityType->getOriginalClass(),
          ];
          $form['overrides'][$x]["current"] = [
            '#markup' => $bundleInfo['class'],
          ];
          if(isset($this->configData()['bundle_class'][$contentEntityTypeId][$bundle]) && $this->configData()['bundle_class'][$contentEntityTypeId][$bundle] == $bundleInfo['class']) {
            $buttonName = [
              'bundle_class',
              $contentEntityTypeId,
              $bundle,
              $bundleInfo['class']
            ];
            $form['overrides'][$x]["remove"] = $this->deleteButton(implode('|', $buttonName));
          }
          $x++;
        }
      }
      //check form classes
      $handlers = $contentEntityType->getHandlerClasses();
      if(isset($handlers['form'])) {
        $forms = $handlers['form'];
        foreach($forms as $operation => $formClass) {
          if(!isset($this->service()->originalFormClasses[$contentEntityTypeId][$operation])) {
            $form['overrides'][$x]["entity_type"] = [
              '#markup' => $contentEntityTypeId . ":" . $operation,
            ];
            $form['overrides'][$x]["override"] = [
              '#markup' => $this->t("Form class"),
            ];
            $form['overrides'][$x]["original"] = [
              '#markup' => 'No class specified',
            ];
            $form['overrides'][$x]["current"] = [
              '#markup' => $forms[$operation],
            ];
            if(isset($this->configData()['form_class'][$contentEntityTypeId][$operation]) && $this->configData()['form_class'][$contentEntityTypeId][$operation] == $forms[$operation]) {
              $buttonName = [
                'form_class',
                $contentEntityTypeId,
                $operation,
                $forms[$operation]
              ];
              $form['overrides'][$x]["remove"] = $this->deleteButton(implode('|', $buttonName));
            }
            $x++;
          }
          elseif($this->service()->originalFormClasses[$contentEntityTypeId][$operation] != $forms[$operation]){
            $form['overrides'][$x]["entity_type"] = [
              '#markup' => $contentEntityTypeId . ":" . $operation,
            ];
            $form['overrides'][$x]["override"] = [
              '#markup' => $this->t("Form class"),
            ];
            $form['overrides'][$x]["original"] = [
              '#markup' => $this->service()->originalFormClasses[$contentEntityTypeId][$operation],
            ];
            $form['overrides'][$x]["current"] = [
              '#markup' => $forms[$operation],
            ];
            if(isset($this->configData()['form_class'][$contentEntityTypeId][$operation]) && $this->configData()['form_class'][$contentEntityTypeId][$operation] == $forms[$operation]) {
              $buttonName = [
                'form_class',
                $contentEntityTypeId,
                $operation,
                $forms[$operation]
              ];
              $form['overrides'][$x]["remove"] = $this->deleteButton(implode('|', $buttonName));
            }
            $x++;
          }
        }
      }
    }
    $keys = array_keys($this->service()->contentEntityTypes);
    $form['overrides']['add']['entity_type'] = [
      '#type' => 'select',
      '#title' => $this->t('Entity type'),
      '#options' => array_combine($keys, $keys),
      '#required' => TRUE,
      '#description' => $this->t('The entity type to override definition.'),
      '#ajax' => [
        'event' => 'change',
        'callback' => [$this , 'ajax_callback'],
      ],
    ];
    $form['overrides']['add']['override'] = [
      '#type' => 'select',
      '#title' => $this->t('Override type'),
      '#options' => [
        'entity_class' => $this->t('Entity class'),
        'bundle_class' => $this->t('Bundle class'),
        'form_class' => $this->t('Form class'),
      ],
      '#required' => TRUE,
      '#description' => $this->t('The type of override.'),
      '#ajax' => [
        'event' => 'change',
        'callback' => [$this , 'ajax_callback'],
      ],
    ];
    $form['overrides']['add']['bundle_operation'] = [
      '#type' => 'select',
      '#title' => $this->t('Bundle or Operation'),
      '#options' => [],
      '#validated' => TRUE,
      '#description' => $this->t('Required if you choose Bundle class or Form class.'),
      '#prefix' => '<div class="bundle-operation-wrapper hidden">',
      '#suffix' => '</div>',
    ];
    $form['overrides']['add']['new_class'] = [
      '#type' => 'textfield',
      '#title' => $this->t('New class'),
      '#required' => TRUE,
      '#description' => $this->t("Must be instance of the current class for that definition."),
    ];
    $form['overrides']['add']['add'] = [
      '#type' => 'submit',
      '#value' => $this->t('Add'),
      '#name' => 'add',
    ];
    $form['actions']['#access'] = FALSE;
    $values = $form_state->getValue('overrides');
    if(isset($values['add'])) {
      $values = $values['add'];
      if(!empty($values['override']) && !empty($values['entity_type'])) {
        if($values['override'] != 'entity_class') {
          $form['overrides']['add']['bundle_operation']['#prefix'] = '<div class="bundle-operation-wrapper">';
          $form['overrides']['add']['bundle_operation']['#required'] = TRUE;
          if($values['override'] == 'bundle_class') {
            $form['overrides']['add']['bundle_operation']['#title'] = $this->t("Select bundle");
            $form['overrides']['add']['bundle_operation']['#options'] = ['' => $this->t("Select")] + $this->service()->getEntityBundles($values['entity_type']);
          }
          elseif($values['override'] == 'form_class'){
            $form['overrides']['add']['bundle_operation']['#title'] = $this->t("Select form operation");
            $form['overrides']['add']['bundle_operation']['#options'] = ['' => $this->t("Select")] + $this->service()->getEntityFormOperations($values['entity_type']);
          }
        }
      }
    }
    return $form;
  }

  public function ajax_callback($form, FormStateInterface $form_state) {
    $response = new AjaxResponse();
    $values = $form_state->getValue('overrides')['add'];
    if(!empty($values['override']) && !empty($values['entity_type'])) {
      if($values['override'] != 'entity_class') {
        $form['overrides']['add']['bundle_operation']['#prefix'] = '<div class="bundle-operation-wrapper">';
        $form['overrides']['add']['bundle_operation']['#required'] = TRUE;
        if($values['override'] == 'bundle_class') {
          $form['overrides']['add']['bundle_operation']['#title'] = $this->t("Select bundle");
          $form['overrides']['add']['bundle_operation']['#options'] = ['' => $this->t("Select")] + $this->service()->getEntityBundles($values['entity_type']);
        }
        elseif($values['override'] == 'form_class'){
          $form['overrides']['add']['bundle_operation']['#title'] = $this->t("Select form operation");
          $form['overrides']['add']['bundle_operation']['#options'] = ['' => $this->t("Select")] + $this->service()->getEntityFormOperations($values['entity_type']);
        }
      }
      else{
        $this->hideBundleField($form);
      }
    }
    else{
      $this->hideBundleField($form);
    }
    $bundleFieldCommand = new ReplaceCommand(".bundle-operation-wrapper" , $form['overrides']['add']['bundle_operation']);
    $response->addCommand($bundleFieldCommand);
    return $response;
  }

  private function hideBundleField(&$form): void{
    $form['overrides']['add']['bundle_operation']['#prefix'] = '<div class="bundle-operation-wrapper hidden">';
    $form['overrides']['add']['bundle_operation']['#options'] = [];
    $form['overrides']['add']['bundle_operation']['#value'] = NULL;
    $form['overrides']['add']['bundle_operation']['#default_value'] = NULL;
    $form['overrides']['add']['bundle_operation']['#required'] = FALSE;
  }

  public function validateForm(array &$form, FormStateInterface $form_state): void  {
    $trigger = $form_state->getTriggeringElement();
    if($trigger['#name'] == 'add') {
      //if form has errors skip validation
      if(count($form_state->getErrors())){
        return;
      }
      //add validation
      $values = $form_state->getValue('overrides')['add'];
      $contentEntityTypeId = $values['entity_type'];
      //verify bundle exists if bundle class override
      if($values['override'] == 'bundle_class'){
        //check bundle not empty
        if(empty($values['bundle_operation'])) {
          $form_state->setErrorByName('overrides][add][bundle_operation' ,$this->t("You must provide a bundle in case of bundle class override."));
          return;
        }
        //validate bundle is valid
        if(empty($this->service()->bundleInfo->getBundleInfo($contentEntityTypeId)[$values['bundle_operation']])){
          $form_state->setErrorByName('overrides][add][bundle_operation' ,$this->t("The bundle @bundle provided is invalid." , ['@bundle' => $values['bundle_operation']]));
          return;
        }
      }
      //verify operation exists if form class override
      if($values['override'] == 'form_class'){
        if(empty($values['bundle_operation'])) {
          $form_state->setErrorByName('overrides][add][bundle_operation' ,$this->t("You must provide an operation in case of form class override."));
          return;
        }
      }
      $result = $this->service()->isValidOverride($values['override'], $contentEntityTypeId, $values['bundle_operation'],$values['new_class']);
      if(!$result->status){
        foreach($result->errors as $error){
          $form_state->setErrorByName('overrides][add' , $error);
        }
      }
    }
    elseif(str_starts_with($trigger['#name'], 'remove|')) {
      $form_state->clearErrors();
      $values = explode('|' , $trigger['#name']);
      if($values[0] == 'entity_class') {
        $contentEntityTypeId = $values[1];
        //check if there is any bundle classes overrides for that entity
        if(isset($this->service()->bundleInfo->getBundleInfo($contentEntityTypeId)[$values[3]]['class'])){
          $form_state->setError($form , $this->t("There is an existing override for bundle @bundle of entity type @entity_type" ,['@entity_type' => $contentEntityTypeId, '@bundle' => $values['bundle_operation']]));
        }
      }
    }
    else{
      $form_state->clearErrors();
    }
  }
  public function submitForm(array &$form, FormStateInterface $form_state){
    $trigger = $form_state->getTriggeringElement();
    $configData = $this->configData();
    if($trigger['#name'] == 'add') {
      $values = $form_state->getValue('overrides')['add'];
      if($values['override'] == 'entity_class') {
        $values['bundle_operation'] = $values['entity_type'];
      }
      $configData[$values['override']][$values['entity_type']][$values['bundle_operation']] = $values['new_class'];
    }
    else{
      //remove
      $data = explode("|", $trigger['#name']);
      unset($configData[$data[1]][$data[2]][$data[3]]);
    }
    $this->cleanConfigData($configData);
    $this->config('entity_classes_override.settings')->setData($configData)->save();
    Cache::getBins()['discovery']->deleteAll();
  }

  public function cleanConfigData(&$configData){
    if(isset($configData['entity_class'])) {
      foreach($configData['entity_class'] as $entityType => $class){
        if(is_array($class) && count($class) == 0){
          unset($configData['entity_class'][$entityType]);
        }
      }
      if(count($configData['entity_class']) == 0){
        unset($configData['entity_class']);
      }
    }
    if(isset($configData['bundle_class'])) {
      foreach($configData['bundle_class'] as $entityType => $bundleClasses){
        if(is_array($bundleClasses) && count($bundleClasses) == 0){
          unset($configData['bundle_class'][$entityType]);
        }
      }
      if(count($configData['bundle_class']) == 0){
        unset($configData['bundle_class']);
      }
    }
    if(isset($configData['form_class'])) {
      foreach($configData['form_class'] as $entityType => $formClasses){
        if(is_array($formClasses) && count($formClasses) == 0){
          unset($configData['form_class'][$entityType]);
        }
      }
      if(count($configData['form_class']) == 0){
        unset($configData['form_class']);
      }
    }
  }

}
