<?php
namespace Drupal\entity_classes_override;
use Drupal\Core\Entity\ContentEntityTypeInterface;
use Drupal\Core\Entity\EntityDisplayRepository;
use Drupal\Core\Entity\EntityTypeBundleInfoInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Plugin\Discovery\AnnotatedClassDiscovery;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use stdClass;

class OverrideService{
  use StringTranslationTrait;

  private EntityTypeManagerInterface $entityTypeManager;
  public array $contentEntityTypes;
  public EntityTypeBundleInfoInterface $bundleInfo;
  public \ArrayObject $namespaces;
  public array $originalFormClasses;
  private EntityDisplayRepository $entityDisplayRepository;

  public function __construct(EntityTypeManagerInterface $entityTypeManager,EntityTypeBundleInfoInterface $entityBundleInfo, \ArrayObject $namespaces,EntityDisplayRepository $entityDisplayRepository){
    $this->entityTypeManager = $entityTypeManager;
    $this->contentEntityTypes = $this->getContentEntityTypes();
    $this->bundleInfo = $entityBundleInfo;
    $this->namespaces = $namespaces;
    $this->originalFormClasses = $this->getOriginalFormClasses();
    $this->entityDisplayRepository = $entityDisplayRepository;
  }

  private function getContentEntityTypes(): array  {
    $entity_types = $this->entityTypeManager->getDefinitions();
    return array_filter($entity_types, function ($entityType) {
      return $entityType instanceof ContentEntityTypeInterface;
    });
  }

  private function getOriginalFormClasses(): array  {
    $discovery = new AnnotatedClassDiscovery('Entity', $this->namespaces, 'Drupal\Core\Entity\Annotation\EntityType');
    $entity_types = $discovery->getDefinitions();
    $entityFormClasses = [];
    foreach($entity_types as $entityTypeId => $entityType) {
      if($entityType instanceof ContentEntityTypeInterface) {
        $handlers = $entityType->getHandlerClasses();
        if(isset($handlers['form'])){
          $entityFormClasses[$entityTypeId] = $handlers['form'];
        }
      }
    }
    return $entityFormClasses;
  }

  public function overrideMessage(bool $result ,array $messages = []):stdClass{
    $output = new stdClass();
    $output->errors = $messages;
    $output->status = $result;
    return $output;
  }

  public function isValidOverride($type,$entity_type_id,$bundle_operation,$new_class):stdClass{
    $messages = [];
    if(!class_exists($new_class)){
      $messages[] = $this->t("The new class @class does not exist." , ['@class' => $new_class]);
      return $this->overrideMessage(FALSE , $messages);
    }
    switch($type){
      case 'entity_class':
        $messages = [];
        if($this->contentEntityTypes[$entity_type_id]->getClass() != $this->contentEntityTypes[$entity_type_id]->getOriginalClass()) {
          $messages[] = $this->t("Entity type @entity_type original class @original_class is already overridden by class @active_class.", [
            '@entity_type' => $entity_type_id,
            '@original_class' => $this->contentEntityTypes[$entity_type_id]->getOriginalClass(),
            '@active_class' => $this->contentEntityTypes[$entity_type_id]->getClass(),
          ]);
        }
        if(!is_subclass_of($new_class, $this->contentEntityTypes[$entity_type_id]->getOriginalClass())){
          $messages[] = $this->t("Can't override entity class for entity type @entity_type, the new class: @new_class must be a subclass of class: @original_class.",[
            '@entity_type' => $entity_type_id,
            '@new_class' => $new_class,
            '@original_class' => $this->contentEntityTypes[$entity_type_id]->getOriginalClass()
          ]);
        }
        foreach($this->bundleInfo->getBundleInfo($entity_type_id) as $bundle => $bundleInfo) {
          if(isset($bundleInfo['class']) && !is_subclass_of($bundleInfo['class'], $new_class)) {
            $messages[] = $this->t("Can't override entity class for entity type @entity_type, an existing override for bundle @bundle class (@bundle_class) is found and it is not a subclass of the new entity class (@new_class).",[
              '@entity_type' => $entity_type_id,
              '@bundle' => $bundleInfo['label'],
              '@bundle_class' => $bundleInfo['class'],
              '@new_class' => $new_class,
            ]);
          }
        }
        break;
      case 'bundle_class':
        if(isset($this->bundleInfo->getBundleInfo($entity_type_id)[$bundle_operation]['class'])){
          $messages[] = $this->t("Can't override bundle class for @bundle of entity type @entity_type, because it is currently overridden by class @active_class.", [
            '@entity_type' => $entity_type_id,
            '@bundle' => $bundle_operation,
            '@active_class' => $this->bundleInfo->getBundleInfo($entity_type_id)[$bundle_operation]['class'],
          ]);
        }
        if(!is_subclass_of($new_class,$this->contentEntityTypes[$entity_type_id]->getClass())) {
          $messages[] = $this->t("Can't override bundle class for @bundle of entity type @entity_type, because the class: @new_class is not a subclass of the active entity class: @entity_class.",[
            '@entity_type' => $entity_type_id,
            '@new_class' => $new_class,
            '@entity_class' => $this->contentEntityTypes[$entity_type_id]->getClass(),
            '@bundle' => $bundle_operation,
          ]);
        }
        break;
      case 'form_class':
        if(isset($this->originalFormClasses[$entity_type_id][$bundle_operation])){
          $handlers = $this->contentEntityTypes[$entity_type_id]->getHandlerClasses();
          if(isset($handlers['form'][$bundle_operation])) {
            if($this->originalFormClasses[$entity_type_id][$bundle_operation] != $handlers['form'][$bundle_operation]){
              $messages[] = $this->t("Can't override form class for operation @operation of entity type @entity_type, because it is currently overridden by class @active_class.", [
                '@operation' => $bundle_operation,
                '@entity_type' => $entity_type_id,
                '@active_class' => $handlers['form'][$bundle_operation],
              ]);
            }
          }
          if(!is_subclass_of($new_class,$this->originalFormClasses[$entity_type_id][$bundle_operation])) {
            $messages[] = $this->t("Can't override form class for operation @operation of entity type @entity_type, because the class: @new_class is not a subclass of the original form class: @form_class.",[
              '@operation' => $bundle_operation,
              '@entity_type' => $entity_type_id,
              '@new_class' => $new_class,
              '@form_class' => $this->originalFormClasses[$entity_type_id][$bundle_operation],
            ]);
          }
        }
        break;
      default:
        $messages = [
          $this->t("Invalid override type @type " , ['@type' => $type]),
        ];
        return $this->overrideMessage(FALSE , $messages);
    }
    if(count($messages)){
      return $this->overrideMessage(FALSE,$messages);
    }
    return $this->overrideMessage(TRUE);
  }

  public function getEntityBundles($entity_type_id): array  {
    $bundles = array_keys($this->bundleInfo->getBundleInfo($entity_type_id));
    return array_combine($bundles, $bundles);
  }

  public function getEntityFormOperations($entity_type_id): array  {
    $handlers = $this->contentEntityTypes[$entity_type_id]->getHandlerClasses();
    $operations = [];
    if(isset($handlers['form'])){
      foreach($handlers['form'] as $form_operation => $handler){
        $operations[$form_operation] = $form_operation;
      }
    }
    $formModes = $this->entityDisplayRepository->getFormModeOptions($entity_type_id);
    foreach($formModes as $form_operation => $form_mode){
      $operations[$form_operation] = $form_operation;
    }
    return $operations;
  }

}
